/************************************************
 * Filename: mastermind.cpp
 * Summary:  Agenda VR3 FLTK MasterMind game
 * Created:  24 February 2001
 * Author:   Kent Dahl ( MenThal@bigfoot.com )
 * Description: A silly little buggy logic game
 ************************************************/

#include "mastermind.hpp"

#include <iostream>
#include <cstring>
#include <cstdlib>
#include <cstdio>
#include <ctime>

#include <sys/time.h>
#include <unistd.h>


class RandomUtil {
public:

    static void randomize()
        {
            timeval t1;
            gettimeofday(&t1,NULL);
            std::srand( t1.tv_usec );
        }

    static unsigned int  rand( unsigned int a_limit ){ 
        unsigned long r = std::rand()-1;
        r &= 0xFFFF;
        r = a_limit*r/0x10000;
        return r; 
    }
};




    DigitConstraints::DigitConstraints( Value a_max, bool a_wrap ) 
        : m_min(0), m_max(a_max), 
          m_wrapAround( a_wrap )
        {}

    Value DigitConstraints::succ( Value a_current ) const
        { 
            return (a_current<m_max ? ++a_current : a_current ); 
        }
   Value DigitConstraints::pred( Value a_current ) const
        { 
            return (a_current>m_min ? --a_current : a_current ); 
        }

    Value DigitConstraints::randomValue() const
        {
            return RandomUtil::rand( m_max - m_min + 1) + m_min;
        }

    bool DigitConstraints::isValid( Value val ) const
        {
            return ( val >= m_min && val <= m_max );
        }




Digit::Digit( Value val, 
           const DigitConstraints &a_constraints,
           ValueListener *a_pListener) 
        : value( val ), 
          m_constraints( a_constraints ),
          m_pListener(a_pListener) { }

    void  Digit::setValue( Value val )
        { value = val; }
    Value Digit::getValue() const
        { return value; }

    void Digit::update()
        { 
            if(m_pListener) m_pListener->valueUpdated( value ); 
        }

    bool Digit::equals( const Digit& a_other ) const
        {
            return value == a_other.value; 
        }

    void Digit::increase()
        {
            value = m_constraints.succ( value );
            update();
        }

    void Digit::decrease()
        {
            value = m_constraints.pred( value );
            update();
        }

    bool Digit::isValid(){ return m_constraints.isValid(value); }

    void Digit::randomize()
        {
            value = m_constraints.randomValue();
        }






//////////////// CODE ////////////////////

    void Code::deleteDigits(){
         // delete the digits
            for( DigitList::iterator it = m_digits.begin();
                 it != m_digits.end();
                 ++it)
            {
                delete (*it);
            }
            m_digits.erase( m_digits.begin(), m_digits.end() );
    }


Code::Code( const CodeConstraints &constraints )
        : m_constraints( constraints ){ }

 Code::~Code()
        {
            deleteDigits();
        }


  bool Code::isValid() const {
        bool result = true;
        DigitList::const_iterator it1    = m_digits.begin();
        DigitList::const_iterator it1end = m_digits.end();

        while( it1 != it1end )
            result &= (*(it1++))->isValid();

        if( m_constraints.m_uniqueDigits ){
            for( DigitList::const_iterator itA = m_digits.begin();
                 itA !=it1end;
                 itA++)
            {
                for( DigitList::const_iterator itB = m_digits.begin();
                     itB !=it1end;
                     itB++)
                {
                    result &= (itA==itB) 
                        || ((*itA)->getValue() != (*itB)->getValue());
                }
            }
           
        }

        return result; 
    }
    bool Code::equals( const Code& a_other )
        { 
            bool result = true;
            DigitList::const_iterator it1    = m_digits.begin();
            DigitList::const_iterator it1end = m_digits.end();
            DigitList::const_iterator it2    = a_other.m_digits.begin();
            DigitList::const_iterator it2end = a_other.m_digits.end();

            while( it1 != it1end && it2 != it2end )
            {
                result &= (*it1)->equals( **it2 );
                ++it1; ++it2;
            }

            
            return result; 
        }
  int  Code::numberOfEqual( const Code& a_other )
        {             
            int result = 0;
            DigitList::const_iterator it1    = m_digits.begin();
            DigitList::const_iterator it1end = m_digits.end();
            DigitList::const_iterator it2    = a_other.m_digits.begin();
            DigitList::const_iterator it2end = a_other.m_digits.end();

            while( it1 != it1end && it2 != it2end )
            {
                if( (*it1)->equals( **it2 ) )
                    result++;

                ++it1; ++it2;
            }
            return result; 
            
        }



int  Code::numberOfExists( const Code& a_other )
{             
    if( m_constraints.m_uniqueDigits )
    {
        return numberOfExists_unique( a_other );
    }
    else
    {
        return numberOfExists_duplicate( a_other );
    }
}


int  Code::numberOfExists_unique( const Code& a_other )
{
    int numValues = m_constraints.m_max - m_constraints.m_min + 1;
    int* values = new int[ numValues ];
    for(int i=0;i<numValues;i++) values[i]=0;
    
    DigitList::const_iterator it1    = m_digits.begin();
    DigitList::const_iterator it1end = m_digits.end();
    DigitList::const_iterator it2    = a_other.m_digits.begin();
    DigitList::const_iterator it2end = a_other.m_digits.end();
    
    int index = 0;
    while( it1 != it1end && it2 != it2end )
    {
        Value v1, v2;
        v1 = (*it1)->getValue();
        v2 = (*it2)->getValue();
        if( v1 == v2 ){
            values[ v1 - m_constraints.m_min ] |= 4; //can't be
        }else{ 
            values[ v1 - m_constraints.m_min ] |= 1;
            values[ v2 - m_constraints.m_min ] |= 2;                
        }                
        ++it1; ++it2; ++index;
    }
    
    int result = 0;
    for (int i=0;i<numValues;i++)
    {
        if(values[i]==3) result++;
        //std::cout<< values[i] <<",";
    }
    delete [] values;
    return result; 
            
}

int  Code::numberOfExists_duplicate( const Code& a_other )
{
    int numDigits = m_constraints.m_numberOfDigits;

    int numExists = 0;
    bool arr[numDigits];

    for(int j=0;j<numDigits;j++) arr[j]=false;

    for(int i=0;i<numDigits;i++)
    {
        // if this isn't a correct hit, continue
        if( !m_digits[i]->equals(*(a_other.m_digits[i])) )
            for(int j=0;j<numDigits;j++ )
        {
            {
                if( !arr[j] 
                    && m_digits[i]->equals(*(a_other.m_digits[j]))
                    )
                {
                    numExists++;
                    arr[j] = true; // digit j in other has been "used"
                    j = numDigits; //ignore rest of digits
                    
                }
            } // j
        } 
        else
        {
            arr[i]=true; // digit used as correct
        }
    } // i



    return numExists;
}

   void Code::randomize()
        {
            if( m_digits.size()==0 )
            {
                for(unsigned int i=0;i<m_constraints.m_numberOfDigits;i++)
                    add( new Digit( m_constraints.m_min, m_constraints ) );
            }

            bool valid = false;
            int tries = 50; // give up eventually; to avoid infinite loop
            while( !valid && tries>0)
            {
                DigitList::const_iterator it1    = m_digits.begin();
                DigitList::const_iterator it1end = m_digits.end();
                while( it1 != it1end )
                {
                    (*it1)->randomize();
                    it1++;
                }
                valid = isValid();
                tries--;
            }
            if( tries==0 ) std::cerr << "Error!";
            if(DEBUG) print(std::cout);

        }

   void Code::print(ostream& out)
        {
            DigitList::const_iterator it1    = m_digits.begin();
            DigitList::const_iterator it1end = m_digits.end();
            out<<"[";
            while( it1 != it1end )
            {
                out << (*(it1++))->getValue()<<", ";
            }        
            out<<"]";

        }

   void Code::add( Digit* a_digit )
        {
            if( a_digit != 0 
                && m_digits.size() < m_constraints.m_numberOfDigits 
                )
                m_digits.push_back( a_digit );
        };






//////////////// GAME ////////////////////




Game::Game( const GameConfig   &a_config,
            ICodeEntryListener &a_listener )
    : m_answer( a_config ),
      
      m_codeEntryListener( a_listener ),
      m_numTries( 0 ),
      m_gameover( false ),
      m_config( a_config )
    
          
        {
            RandomUtil::randomize();
            m_answer.randomize();
            m_codeEntryListener.gameBegins();
        }
     Game::~Game(){}

     void Game::enter( const Code &a_entry )
        {
            // if in "zombie" mode after a game over
            if( m_gameover ) return;
            

            // invalid entries ignored
            if (!a_entry.isValid())
            {
                m_codeEntryListener.invalidEntry();
                return;
            }

            m_numTries++;

            if ( m_answer.equals( a_entry ) )
            {
                m_codeEntryListener.playerWon( m_numTries );
                m_gameover = true;
                m_codeEntryListener.gameOver();
                return;
            }

            int numCorrect = m_answer.numberOfEqual(  a_entry );
            int numExists  = m_answer.numberOfExists( a_entry );
            CodeEntryResult result( numCorrect, numExists );
            m_codeEntryListener.entered( result );

            if( m_numTries>=m_config.m_maxTries )
                m_codeEntryListener.gameOver();
                
        }

     void Game::restart()
        {
            m_answer.randomize();
            m_numTries = 0;
            m_gameover = false;
            m_codeEntryListener.gameBegins();
        }
    




////////////////////////////////

