/************************************************
 * Filename: mastermind.cpp
 * Summary:  Agenda VR3 FLTK MasterMind game
 * Created:  24 February 2001
 * Author:   Kent Dahl ( MenThal@bigfoot.com )
 * Description: A silly little buggy logic game
 ************************************************/

#include <FL/Fl.H>
#include <FL/Fl_Window.H>
#include <FL/Fl_Box.H>
#include <FL/Fl_Button.H>
#include <FL/Fl_Browser.H>
#include <FL/Fl_Pack.H>
#include <FL/Fl_Menu_Button.H>

#include <iostream>
#include <cstring>
#include <cstdlib>
#include <ctime>
#include <cstdio>
#include <vector>

#include "mastermind.hpp"

#define XSIZE 160
#define YSIZE 240


static bool MASTERMIND_CHEAT = false;
//static bool MASTERMIND_DEBUG = false;







class MMDigit : public Digit, public ValueListener
{
    char *m_text;
    Fl_Widget *m_widget;
    const char *m_digitRepresentation;
public:

    static char*  DIGITS;
    
public:
    MMDigit( Value val, const DigitConstraints &a_constraints,
             //ValueListener *a_pListener = 0,
             const char *a_digitRep = 0 )
        : Digit( val, a_constraints, this),
          m_widget( 0 ),
          m_digitRepresentation( a_digitRep )
        {
            m_text = new char[2];
            m_text[0] = m_digitRepresentation[value];
            m_text[1] = 0; 
        }
    
    char* getText(){ return m_text; }
    

    void valueUpdated( Value value )
        { 
            char c = m_digitRepresentation[value];
            m_text[0] = c;   
            
            if(DEBUG)std::cout<< "Value:"<<value<<" Ch:"<<c<<"\n";
            
            if(m_widget){
                m_widget->label( m_text);
                if(DEBUG)std::cout<< m_text;
                m_widget->redraw();
            }
        }
    
    void setWidget( Fl_Widget *a_w ){ m_widget = a_w; }
    
};

char* MMDigit::DIGITS = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ!#%&()=?+-*/";

class MMCode : public Code
{
    const char *m_digitRepresentation;
public:
    MMCode( const CodeConstraints &constraints,
            const char *digit_rep = MMDigit::DIGITS )
        : Code( constraints) ,
          m_digitRepresentation( digit_rep )
        { }


    char *createString(){
        char *ch = new char[ m_digits.size() + 1 ];
        int index = 0;
        for( DigitList::const_iterator it = m_digits.begin();
             it != m_digits.end();
             ++it )
        {
            ch[index++] = m_digitRepresentation[(*it)->getValue()];
        }
        ch[index] = '\0';

        return ch;
    }

};




class MMGameConfig : public GameConfig 
{
public:
    const char *m_name;
    const char *m_digitRepresentation;
    MMGameConfig( const char *a_name = "(config)",
                  const char *a_digRep = MMDigit::DIGITS,
                  int a_maxTries = 10,
                  int a_numDigits = 4,
                  int a_numValues = 6,
                  bool a_uniqueDigits = true )
        : GameConfig( a_maxTries, a_numDigits, a_numValues, a_uniqueDigits ), 
          m_name( a_name ),
          m_digitRepresentation( a_digRep )
        {
        }
};


class MMGame : public CodeEntryListener_Base {
protected:
    MMGameConfig *m_pConfig;
    Game m_game;
    MMCode m_entry;

    Fl_Browser *m_feedbackWidget;
public:
    static char *FEEDBACK_STRING;


    ~MMGame(){ delete m_pConfig; }
        

    static void butt_cb(Fl_Widget *w, void *data){
        if(DEBUG)std::cout<<"Button: "<<w<<", "<<data<<"\n";
    }

    static void buttonUp(Fl_Widget *w, void *data){
        MMDigit* digit = static_cast<MMDigit*>(data);
        digit->increase();
    }
    static void buttonDown(Fl_Widget *w, void *data){
        MMDigit* digit = static_cast<MMDigit*>(data);
        digit->decrease();
    }
    static void buttonEnter(Fl_Widget *w, void *data){
        MMGame* game = static_cast<MMGame*>(data);
        game->m_game.enter( game->m_entry );
    }


    void start(){
        feedback("Good luck!");

        char chars[25];
        char *ch = chars; //new char[25];
        std::sprintf(ch, "You have %i tries.", m_pConfig->m_maxTries );
        feedback(ch);

        if ( m_pConfig->m_uniqueDigits )
            feedback("All digits unique.");
        else 
            feedback("Duplicates allowed.");

        std::sprintf(ch, "%i digits long.", 
                     m_pConfig->m_numberOfDigits );
        feedback(ch);

        std::sprintf(ch, "%i characters,", 
                     m_pConfig->m_max - m_pConfig->m_min +1  );
        feedback(ch);

        feedback("The code is:");
        feedback("____________");

     
        if(MASTERMIND_CHEAT)
        {
            std::cout<<"Code:";
            m_game.print( std::cout );
        }



    } // 
    void exit(){  } // 


    MMGame( MMGameConfig *a_pConfig )
        : m_pConfig( a_pConfig ),
          m_game( *a_pConfig, *this ),
          m_entry( *a_pConfig, a_pConfig->m_digitRepresentation )
        {
            if(DEBUG)std::cout<<"Game setup\n";
    
            unsigned int numDigits = m_pConfig->m_numberOfDigits;
            int xsize = 100, ysize=150;

            int edity=100;
            int editsize = 15;

            xsize = numDigits*editsize + 3*2 + 2*editsize+ 3;
            if (xsize>XSIZE-10){
                xsize = XSIZE-10;
                editsize = (xsize-3*4) / (numDigits+2);
            }

            if(DEBUG)std::cout<<"xsize="<<xsize<<"\n";
            
            // Create Window
            Fl_Window *window = new Fl_Window(xsize,ysize,"MasterMind");
        

        // make up and down buttons;

    


        for( unsigned int i = 0; i<numDigits; i++){
            
            MMDigit *digit 
                = new MMDigit( m_pConfig->m_min,
                               *m_pConfig,
                               m_pConfig->m_digitRepresentation);
            m_entry.add( digit );

             Fl_Button *but;
             but = new Fl_Button(3+editsize*i,edity,
                                 editsize,editsize,
                                 "@#2<-");
             but->labeltype(FL_SYMBOL_LABEL);
             but->callback( buttonUp, digit );


             Fl_Box *box 
                 = new Fl_Box(FL_DOWN_BOX,
                              3+editsize*i,edity+1*editsize,
                              editsize,editsize,
                              digit->getText());
             
             digit->setWidget( box );
             
             but = new Fl_Button(3+editsize*i,edity+2*editsize,
                                 editsize,editsize,
                                 "@#8<-");
             but->labeltype(FL_SYMBOL_LABEL);
             but->callback( buttonDown, digit );
             

        }



        Fl_Button *but 
            = new Fl_Button( 3+3+editsize*numDigits, edity, 
                             editsize*2,editsize*2,
                             "@#returnarrow");
        but->labeltype(FL_SYMBOL_LABEL);
        but->callback( buttonEnter, this );
        
        //m_enterButton = but;
/*
        Fl_Menu_Button *mbut
            = new Fl_Menu_Button(

                3+3+editsize*numDigits, edity+editsize*2, 
                editsize*2,editsize,
                ""
                );
        mbut->labeltype(FL_SYMBOL_LABEL);
        mbut->type( 0 );
        mbut->add( "Close" );
        mbut->add( "Restart" );
        //mbut->add( "" );
   */
    

        Fl_Browser *gr 
            = new Fl_Browser(1,0,xsize-5,edity-5,"");
        gr->textsize(8);

        m_feedbackWidget = gr;






        window->end();

        // Display the window
        window->show();

        }

    virtual void gameBegins()
        {
            if(DEBUG)std::cout<<"CodeEntryListener: gameBegins\n"; 
            //feedback("Good luck.\0");
        }
    
    virtual void playerWon( int a_numTries )
        { 
            if(DEBUG)std::cout<<"CodeEntryListener: playerWo \n"; 

            char ch[25];
            sprintf(ch, "Using %i tries.", a_numTries );
            feedback(ch);

            feedback("You did it!!!");
        }

    virtual void entered( CodeEntryResult result )
        { 
            if(DEBUG)std::cout<<"CodeEntryListener: entered\n"; 
            char *ch1 = m_entry.createString();
            int fbs = std::strlen(FEEDBACK_STRING);
            int cs = std::strlen(ch1);
            char *ch2 = new char[ fbs + cs + 1 ];
            std::strcpy( ch2, ch1 );
            std::strcpy( ch2 + cs, FEEDBACK_STRING);

            ch2[fbs + cs] = '\0';

            char *p = std::strchr( ch2, '#');
            *p =(char)('0'+result.m_numCorrect);
            
            p = std::strchr( p, '#');
            *p =(char)('0'+result.m_numExists  );


            feedback( ch2 );
            delete [] ch1;
            delete [] ch2;
        }

    virtual void gameOver()
        {
            if(DEBUG)std::cout<<"CodeEntryListener:  gameOver\n";
            feedback("Game over!!!");
        }

    virtual void invalidEntry()
        { 
            if(DEBUG)std::cout<<"CodeEntryListener: invalidEntry\n"; 
        }
    void feedback( char * ch ){
        // put it at the top
        m_feedbackWidget->insert(0, ch, ch );
        m_feedbackWidget->redraw();
    }

};

char *MMGame::FEEDBACK_STRING = " - (#,#) \0x00";





class MMStartScreen 
{
    typedef MMGameConfig* (*ConfigCreatorFunc)();

    static int           NUM_CONFIGS;
    static MMGameConfig  CONFIG_ARRAY[];

public:
    static void newGame(Fl_Widget *w, void *data){
        if(DEBUG)std::cout<<"New game\n";
        MMGameConfig *conf;
        if(data){
            conf = static_cast<MMGameConfig*>(data);
            conf = new MMGameConfig(*conf);
        }else{
            conf = new MMGameConfig(CONFIG_ARRAY[0]);
        }

        MMGame *game = new MMGame( conf );
        game->start();
        
    }

    static MMGameConfig *createNormalConfig(){ 
        return new MMGameConfig(); //"abcdefghijklmnopqrstuvwxyz0123456789"); 
    }

    static MMGameConfig *createAlfaConfig(){ 
        return new MMGameConfig(); //"abcdefghijklmnopqrstuvwxyz0123456789"); 
    }



    void setup( MMGameConfig *a_pUserDefConfig = 0  )
        {
            Fl_Window *window = new Fl_Window(100,130,"MasterMind");


            Fl_Box *label;
            label = new Fl_Box( 0,5,100,25, "MasterMind" );
            label->labelsize(16);

            Fl_Button *but;
            but = new Fl_Button( 10,35,80,25,"Play NOW");
            but->callback( newGame, &CONFIG_ARRAY[1] );

            Fl_Menu_Button *mbut
                = new Fl_Menu_Button(
                    10,65,80,25,
                    "New..."
                    );
            mbut->labeltype(FL_SYMBOL_LABEL);
            mbut->type( 0 );
            
            for( int i = 0;i<NUM_CONFIGS; i++)
            {
                mbut->add( CONFIG_ARRAY[i].m_name, "", 
                           newGame, &CONFIG_ARRAY[i] );
            }


            if( a_pUserDefConfig )
                mbut->add( "User defined", "", newGame, a_pUserDefConfig );

            label = new Fl_Box( 0,95,100,25, "by Kent Dahl" );
            label->labelsize(8);

            window->end();
            // Display the window
            window->show();

        }



};

int          MMStartScreen::NUM_CONFIGS = 3;
MMGameConfig MMStartScreen::CONFIG_ARRAY[] = {
    MMGameConfig("Easy",   "0123456789", 15, 4, 6, true ),
    MMGameConfig("Normal", "0123456789", 10, 4, 6, true ),
    MMGameConfig("Hard",   "0123456789", 15, 5, 8, true ),
    MMGameConfig("Funky!", "aBcDeFgHiJkLmNoPqRsTuV")
};

/*
 * "classic" 4, 6, true
 * 
*/




int main(int argc, char **argv) 
{

    bool createUserDefined = false;
    unsigned int numTries = 10;
    unsigned int numDigits = 4;
    unsigned int numValues = 6;
    char *charRep = MMDigit::DIGITS;
    bool unique = true;
    
    MMGameConfig *pUserDefConfig = 0;

    // command line arguments
    if(argc>1)
        for(int i = 1;i<argc;i++)
        {
            if( std::strcmp( argv[i],"-cheat")==0 ) 
                MASTERMIND_CHEAT = true;

            if( std::strcmp( argv[i],"-tries")==0 && argc>i+1 )
            {
                numTries = atoi(argv[i+1]);
                createUserDefined = true;
            }
            if( std::strcmp( argv[i],"-digits")==0 && argc>i+1 )
            {
                numDigits = atoi(argv[i+1]);
                createUserDefined = true;
            }
            if( std::strcmp( argv[i],"-values")==0 && argc>i+1 )
            {
                numValues = atoi(argv[i+1]);
                createUserDefined = true;
            }
            if( std::strcmp( argv[i],"-charrep")==0 && argc>i+1 )
            {
                if( std::strlen( argv[i+1] )>= numValues )
                    charRep = argv[i+1];
                createUserDefined = true;
            }

            if( std::strcmp( argv[i],"-duplicate")==0  )
            {
                unique = false;
                createUserDefined = true;
            }

        }

    if( createUserDefined )
        pUserDefConfig = new MMGameConfig(
            "User defined", charRep,
            numTries, numDigits, numValues, unique );

    MMStartScreen start;
    start.setup( pUserDefConfig );


    // Start the main loop
    return Fl::run();
}


///////////////////////////////////////
